//--------------------------------------------------------------------------------------
// File: Tutorial02.cpp
//
// Setting up a Direct3D 10 Device and Swap Chain
//
//--------------------------------------------------------------------------------------
#include <windows.h>
#include <d3d10.h>
#include <d3dx10.h>

//--------------------------------------------------------------------------------------
// Global Variables
//--------------------------------------------------------------------------------------
HINSTANCE               g_hInst = NULL;
HWND                    g_hWnd = NULL;

D3D10_DRIVER_TYPE       g_driverType = D3D10_DRIVER_TYPE_NULL;
ID3D10Device*           g_pd3dDevice = NULL;
IDXGISwapChain*         g_pSwapChain = NULL;
ID3D10RenderTargetView* g_pRenderTargetView = NULL;


//--------------------------------------------------------------------------------------
// Forward declarations
//--------------------------------------------------------------------------------------
HRESULT InitWindow( HINSTANCE hInstance, int nCmdShow );
HRESULT InitDevice();
void CleanupDevice();
LRESULT CALLBACK    WndProc( HWND, UINT, WPARAM, LPARAM );
void Render();


//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
	if( FAILED( InitWindow( hInstance, nCmdShow ) ) )
		return 0;

	if( FAILED( InitDevice() ) )
	{
		CleanupDevice();
		return 0;
	}

	// Main message loop

	// We change the message loop by replacing GetMessage() with PeekMessage(). In contrast to
	// GetMessage(), PeekMessage returns immediately even if the message loop is empty
	MSG msg = {0};
	while( WM_QUIT != msg.message )
	{
		if( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )
		{
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
		else
		{
			// if there are no events pending, we render the scene
			Render();
		}
	}

	CleanupDevice();

	return ( int )msg.wParam;
}


//--------------------------------------------------------------------------------------
// Register class and create window
//--------------------------------------------------------------------------------------
HRESULT InitWindow( HINSTANCE hInstance, int nCmdShow )
{
	// Register class
	WNDCLASSEX wcex;

	// clear out the window class for use
	ZeroMemory(&wcex, sizeof(WNDCLASSEX)); // never forget to zero out, if you do not set all the values!

	// fill the window class struct
	wcex.cbSize			= sizeof( WNDCLASSEX );
	wcex.style			= CS_HREDRAW | CS_VREDRAW; // Redraw when width or height changes
	wcex.lpfnWndProc	= WndProc; // function pointer 
	wcex.hInstance		= hInstance; // Instance handle: important to find window class (CreateWindow)
	wcex.hCursor		= LoadCursor( NULL, IDC_ARROW ); // shape of the cursor
	wcex.hbrBackground	= ( HBRUSH )( COLOR_WINDOW );    // background color
	wcex.lpszClassName	= L"TutorialWindowClass"; // Class name: important to find window class (CreateWindow)

	if( !RegisterClassEx( &wcex ) )
		return E_FAIL;

	// Create window
	g_hInst = hInstance;
	RECT rc = { 0, 0, 640, 480 };
	AdjustWindowRect( &rc, WS_OVERLAPPEDWINDOW, FALSE ); // adjust to match client rect
	g_hWnd = CreateWindow( L"TutorialWindowClass", L"Direct3D 10 Tutorial 0: Setting Up Window", WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT, rc.right - rc.left, rc.bottom - rc.top, NULL, NULL, hInstance,
		NULL );
	if( !g_hWnd )
		return E_FAIL;

	ShowWindow( g_hWnd, nCmdShow );

	return S_OK;
}

//--------------------------------------------------------------------------------------
// Called every time the application receives a message
//--------------------------------------------------------------------------------------
LRESULT CALLBACK WndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
	PAINTSTRUCT ps;
	HDC hdc;

	switch( message )
	{
	case WM_PAINT:
		hdc = BeginPaint( hWnd, &ps );
		EndPaint( hWnd, &ps );
		break;

	case WM_DESTROY:
		PostQuitMessage( 0 );
		break;

	default:
		return DefWindowProc( hWnd, message, wParam, lParam );
	}

	return 0;
}


//--------------------------------------------------------------------------------------
// Create Direct3D device and swap chain
//--------------------------------------------------------------------------------------
HRESULT InitDevice()
{
	HRESULT hr = S_OK;

	// get the size of the window's client rect
	RECT rc;
	GetClientRect( g_hWnd, &rc );
	UINT width = rc.right - rc.left;
	UINT height = rc.bottom - rc.top;


	UINT createDeviceFlags = 0;
#ifdef _DEBUG
	createDeviceFlags |= D3D10_CREATE_DEVICE_DEBUG;
#endif


	// find an type of dx10 driver supported by the hardware
	D3D10_DRIVER_TYPE driverTypes[] =
	{
		D3D10_DRIVER_TYPE_HARDWARE,
		D3D10_DRIVER_TYPE_REFERENCE,
	};
	UINT numDriverTypes = sizeof( driverTypes ) / sizeof( driverTypes[0] );

	// fill the swap chain descriptor
	DXGI_SWAP_CHAIN_DESC sd; 
	
	ZeroMemory( &sd, sizeof( sd ) );
	// number of buffers
	sd.BufferCount       = 1;

	// buffer descriptor
	sd.BufferDesc.Width  = width; 
	sd.BufferDesc.Height = height;
	sd.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
	sd.BufferDesc.RefreshRate.Numerator = 60;
	sd.BufferDesc.RefreshRate.Denominator = 1;

	// buffer usage: use as a render target (might as well be used as a texture)
	sd.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;

	// handle to output window that will show the front buffer
	sd.OutputWindow = g_hWnd;

	// multisampling support (turned off for now)
	sd.SampleDesc.Count = 1;
	sd.SampleDesc.Quality = 0;

	// windowed or fullscreen
	sd.Windowed = TRUE;

	// try to create a swap chain for a specific driver type
	for( UINT driverTypeIndex = 0; driverTypeIndex < numDriverTypes; driverTypeIndex++ )
	{
		g_driverType = driverTypes[driverTypeIndex];
		hr = D3D10CreateDeviceAndSwapChain( 
				NULL,				// [in] you can use a specific graphics device here, NULL means use the first one
				g_driverType,		// [in] driver type
				NULL,				// [in] RESERVED: Used in future versions to hand over a software rasterizer
				createDeviceFlags,	// [in] create flags
				D3D10_SDK_VERSION,	// [in] SDK version
				&sd,				// [in] swap chain descriptor (see above)
				&g_pSwapChain,		// [out] pointer to swap chain interface
				&g_pd3dDevice );	// [out] pointer to D3D device interface
		if( SUCCEEDED( hr ) )
			break; // break out of the for loop if creation succeeded
	}
	if( FAILED( hr ) )
		return hr;

	// Resources is just raw memory on the device
	// A resource view allows you to specify the use of the memory (as a frame buffer, as a texture etc)
	// We need a render target view to get access to the back buffer in the swap chain
	//
	// a chunk of memory in D3D is referred to using the Texture2D interface
	//
	ID3D10Texture2D* pBackBuffer;
	hr = g_pSwapChain->GetBuffer( 0, __uuidof( ID3D10Texture2D ), ( LPVOID* )&pBackBuffer );
	if( FAILED( hr ) )
		return hr;

	// Create a render target view onto the memory
	//
	hr = g_pd3dDevice->CreateRenderTargetView( pBackBuffer, NULL, &g_pRenderTargetView );
	pBackBuffer->Release();
	if( FAILED( hr ) )
		return hr;

	g_pd3dDevice->OMSetRenderTargets(
		1, // number of render targets, (for use with multiple render targets)
		&g_pRenderTargetView, // array of render target views to be bound
		NULL // pointer to Depth/StencilView
	);

	// Setup the viewport
	D3D10_VIEWPORT vp;
	vp.Width = width;
	vp.Height = height;
	vp.MinDepth = 0.0f;
	vp.MaxDepth = 1.0f;
	vp.TopLeftX = 0;
	vp.TopLeftY = 0;
	g_pd3dDevice->RSSetViewports( 1, &vp );

	return S_OK;
}


//--------------------------------------------------------------------------------------
// Render the frame
//--------------------------------------------------------------------------------------
void Render()
{
	// Just clear the backbuffer
	float ClearColor[4] = { 0.5f, 0.5, 1.0f, 1.0f }; //red,green,blue,alpha
	g_pd3dDevice->ClearRenderTargetView( g_pRenderTargetView, ClearColor );
	g_pSwapChain->Present( 0, 0 );
}


//--------------------------------------------------------------------------------------
// Clean up the objects we've created
//--------------------------------------------------------------------------------------
void CleanupDevice()
{
	if( g_pd3dDevice ) g_pd3dDevice->ClearState();

	if( g_pRenderTargetView ) g_pRenderTargetView->Release();
	if( g_pSwapChain ) g_pSwapChain->Release();
	if( g_pd3dDevice ) g_pd3dDevice->Release();
}


